#! ruby -Ks
# shift-jis: "-Ks"
# euc      : "-Ke"   Please change Top line "#! ruby -Ks"
# utf      : "-Ku"             depending on the situation
# Version 4.0.2: Improvement for processing time
#                  and document score
# Author: Hiroshi Echizen'ya
# Date: 01/06/2013
##########################################################

#
#  Learn
#
class Learn
  def initialize(name, re_op, re, len_op, len)
    @name = name
    if re_op == nil
      @rec_op = "-r"
      @rec = 0.1
      @len_op = "-l"
      @len = 1.2
    else
      @rec_op = re_op
      @rec = re
      @len_op = len_op
      @len = len
    end
    @all_sent = Array.new
    @all_origin_sent = Array.new
    @unit_origin_sent = Array.new
    @unit_sent = Array.new
    @wk_sent = Array.new
    @unit_sent = Array.new
    @unit_origin_sent = Array.new
    @max_route_num = 0
    @route_num = 0
    @all_ip_score = 0.0
    @limit_route_num = 1000000
  end

  #
  #  delete a line feed code
  #
  def chop_term(line)
    line.sub(/(?:\r|\r\n|\n)\z/, "")
  end

  #
  #  delete all line feed codes
  #
  def chop_term_loop(line)
    line_tmp = chop_term(line)
    while line_tmp != line
      line = line_tmp
      line_tmp = chop_term(line)
    end
    line
  end

  #
  #  kutou_token_loop
  #
  def kutou_token_loop(line)
    line_tmp = kutou_token(line)
    while line_tmp != line
      line = line_tmp
      line_tmp = kutou_token(line)
    end
    line
  end
  
  #
  # kutou_token
  #
  def kutou_token(line)
    if line =~ /^<\/seg>$/
      return line
    end
    line = line.gsub(/([\.])([^ ^\.])/) { $1+' '+$2 }    # word. -> word .
    line = line.gsub(/([^ ^\.])([\.])/) { $1+' '+$2 }    # .word -> . word
    line = line.gsub(/([\.])([ ])([\.])/) { $1+$3 } # . . -> ..: Delete Space
    line = line.gsub(/([0-9])([ ])([\.])([ ])([0-9])/) { $1+$3+$5 } # num . num -> num.num: 0 . 3 -> 0.3: Delete Space
    line = line.gsub(/([^ ])([ ])([\.])([ ])([^ ^0-9])([ ])([\.])/) { $1+$3+$5+$7 } # * . * . -> *.*. : e . g . -> e.g.: Delete Space
    line = line.gsub(/([^ ^\"])([\"])/) { $1+' '+$2 }          # word" -> word "
    line = line.gsub(/([\"])([^ ^\"])/) { $1+' '+$2 }          # "word -> " word
    line = line.gsub(/([^ ^\(])([\(])/) { $1+' '+$2 }          # word( -> word (
    line = line.gsub(/([\)])([^ ^\)])/) { $1+' '+$2 }          # )word -> ) word
    line = line.gsub(/([\(])([^ ^\(])/) { $1+' '+$2 }          # (word -> ( word 
    line = line.gsub(/([^ ^\)])([\)])/) { $1+' '+$2 }          # word) -> word )
    line = line.gsub(/([\-])([^ ^\-])/) { $1+' '+$2 }          # -word -> - word
    line = line.gsub(/([^ ^\-])([\-])/) { $1+' '+$2 }          # word- -> word -
    line = line.gsub(/([_])([^ ^_])/) { $1+' '+$2 }            # _word -> - word
    line = line.gsub(/([^ ^_])([_])/) { $1+' '+$2 }            # word_ -> word _
    line = line.gsub(/([^ ^\/])([\/])/) { $1+' '+$2 }          # word/ -> word /
    line = line.gsub(/([\/])([^ ^\/])/) { $1+' '+$2 }          # /word -> / word
    line = line.gsub(/([^ ^:])([:])/) { $1+' '+$2 }            # word: -> word :
    line = line.gsub(/([:])([^ ^:])/) { $1+' '+$2 }            # :word -> : word
    line = line.gsub(/([^ ^;])([;])/) { $1+' '+$2 }            # word; -> word ;
    line = line.gsub(/([;])([^ ^;])/) { $1+' '+$2 }            # ;word -> ; word
    line = line.gsub(/([^ ])([,])([ ])/) { $1+' '+$2+$3 }      # word,  -> word , , 1,234 -> 1,234
    line
  end

  #
  #  para_check
  #
  def para_check()
    if @rec_op != "-r"
      print("Command Line Error: \"-r\" option", "\n")
      return -1
    end
    if @rec > 1.0 || @rec < 0.0
      print("Input Error: Please input between 0.0 and 1.0 as \"-r\" option", "\n")
      return -1
    end
    if @len_op != "-l"
      print("Command Line Error: \"-l\" option", "\n")
      return -1
    end
    if @len < 1.0
      print("Input Error: Please input over 1.0 as \"-l\" option", "\n")
      return -1
    end
    if FileTest.exist?(@name) == false
      print("No such file - ", "\"", @name, "\"",  "\n")
      return -1
    end
  end
  
  #
  #  read_data
  #
  def read_data()
    @all_sent_num = 0
    @all_origin_sent = []
    @all_sent = []
    @wk_sent = [[]]
    unit_origin_sent_wk = []
    unit_sent_wk = []
    unit_sent_cnt = 0
    start_flg = 0
    File.open(@name, "r"){ |input_file|
      start_flg = 0
      input_file.each { |line|
        line = kutou_token_loop(line)
        line = chop_term_loop(line)
        if line =~ /^<seg id=/
          if start_flg == 0
            start_flg = 1
            unit_sent_cnt = 0
#printf("%s\n", line)
          else
            print("Format Error!! Not Found <\/seg>\n")
            return -1
          end
          next
        end
        if line =~ /^<\/seg>$/
          @all_origin_sent << unit_origin_sent_wk
          @all_sent << unit_sent_wk
          start_flg = 0
          unit_sent_cnt = 0
          unit_origin_sent_wk = []
          unit_sent_wk = []
          @all_sent_num += 1
#printf("%s\n\n", line)
          next
        end
        if start_flg == 1
          unit_origin_sent_wk[unit_sent_cnt] = line.split(/ /)
          unit_origin_sent_wk[unit_sent_cnt].delete("")
          line = line.downcase
          unit_sent_wk[unit_sent_cnt] = line.split(/ /)
          unit_sent_wk[unit_sent_cnt].delete("")
          
=begin
i=0
  while i<unit_sent_wk[unit_sent_cnt].length
  printf("%s ", unit_sent_wk[unit_sent_cnt][i])
  i+=1
end
printf("\n")
=end
          if unit_sent_cnt == 0
            @wk_sent[@all_sent_num] = line.split(/ /)
            @wk_sent[@all_sent_num].delete("")
          end
          unit_sent_cnt += 1
        end
        if start_flg == 1
          next
        end
      }
    }
    if @all_sent_num == 0
      print("Format Error!! Not Found <seg id=*>\n")
      return -1
    end
  end
  
  #
  #  impact_process
  #
  def impact_process
    if para_check() == -1
      return
    end
    if read_data() == -1
      return
    end
    all_sent_cnt = 0
    while all_sent_cnt < @all_sent_num
      impact_score(all_sent_cnt)
      all_sent_cnt += 1
    end

    all_ip = @all_ip_score / @all_sent.length
  printf("test set=%.4f\n", all_ip)
  end
  
  #
  # create_unit
  #
  def create_unit(all_sent_cnt)
    wk_sent = Array.new
    wk_unit_sent = Array.new
    sent_cnt = 0
    while sent_cnt < @all_sent[all_sent_cnt].length
      wd_cnt = 0
      wk_unit_sent = []
      while wd_cnt < @all_sent[all_sent_cnt][sent_cnt].length
        wk_unit_sent[wd_cnt] = @all_sent[all_sent_cnt][sent_cnt][wd_cnt]
        wd_cnt += 1
      end
      wk_sent << wk_unit_sent
      sent_cnt += 1
    end
    @unit_sent = []
    @unit_sent = wk_sent
  end
  
  #
  # create_origin_unit
  #
  def create_origin_unit(all_sent_cnt)
    wk_sent = Array.new
    wk_unit_sent = Array.new
    sent_cnt = 0
    while sent_cnt < @all_origin_sent[all_sent_cnt].length
      wd_cnt = 0
      wk_unit_sent = []
      while wd_cnt < @all_origin_sent[all_sent_cnt][sent_cnt].length
        wk_unit_sent[wd_cnt] = @all_origin_sent[all_sent_cnt][sent_cnt][wd_cnt]
        wd_cnt += 1
      end
      wk_sent << wk_unit_sent
      sent_cnt += 1
    end
    @unit_origin_sent = []
    @unit_origin_sent = wk_sent
  end

  #
  #  claculate impact score
  #
  def impact_score(all_sent_cnt)
    create_unit(all_sent_cnt)
    @unit_origin_sent = []
    create_origin_unit(all_sent_cnt)
    @unit_sent_num = @unit_sent.length
    max_rip_score = 0
    max_pip_score = 0
    max_rip_sent = 0
    max_pip_sent = 0
    max_rip_norm = 0.0
    max_pip_norm = 0.0
    max_rip_lcs_score = 0.0
    max_pip_lcs_score = 0.0
    
    unit_sent_cnt=1
    while unit_sent_cnt < @unit_sent_num
      all_lcs_score = get_all_cp_score(0, unit_sent_cnt)
      
      create_unit(all_sent_cnt)
                                 # calculate Pip score
      pip_norm_score = @unit_sent[0].length ** @len
      pip_score = (all_lcs_score / pip_norm_score) ** (1.0/@len)
      if pip_score >= max_pip_score
        max_pip_score = pip_score
        max_pip_norm = pip_norm_score
        max_pip_lcs_score = all_lcs_score
      end
                                 # calculate Rip score
      rip_norm_score = @unit_sent[unit_sent_cnt].length ** @len
      rip_score = (all_lcs_score / rip_norm_score) ** (1.0/@len)
      if rip_score >= max_rip_score
        max_rip_score = rip_score
        max_rip_norm = rip_norm_score
        max_rip_lcs_score = all_lcs_score
      end

      @unit_sent[0] = @wk_sent[all_sent_cnt].dup
      unit_sent_cnt += 1
    end

    ip_score = 0
    if all_lcs_score != 0
      beta = max_pip_score / max_rip_score
      ip_score = ((1 + beta ** 2) * max_rip_score * max_pip_score) / (max_rip_score + (beta ** 2 * max_pip_score))
    end
    
# printf("max_rip = %.4f max_pip = %.4f\n", max_rip_lcs_score, max_pip_lcs_score)
    printf("%.4f\n", ip_score)

    @all_ip_score += ip_score
  end

  #
  #  get_all_cp_score
  #
  def get_all_cp_score(no1, no2)
    cp_score = 0
    
    @list = [[]]
    i = 0
    while i < @unit_sent[no2].length+1
      @list[i] = [0]
      i += 1
    end
    
    j = 1
    while j < @unit_sent[no1].length+1
      @list[0][j] = 0
      j += 1
    end
    
    i = 0
    loop {
      unit_score = 0
      unit_score = get_lcs_route(no1, no2)
      if unit_score == 0
        break
      else
        cp_score += (unit_score * (@rec ** i))
        i += 1
      end
    }
    cp_score
  end

  #
  #  cal_cps_score
  #
  def cal_cps_score(com_no1, com_no2, no1, no2, flg)
    com_cnt = 0
    cps_score = 0
    len1 = @unit_sent[no1].length
    len2 = @unit_sent[no2].length
    while com_cnt < com_no1.length
      if flg == 1
        pos_no1 = com_no1[com_cnt][0]
        pos_no2 = com_no2[com_cnt][0]
        diff = ((pos_no1+1).to_f/len1)-((pos_no2+1).to_f/len2)
        diff = diff.abs
        pst_score = 1.0 - diff
        lcs_score = com_no1[com_cnt].length.to_f ** @len
        cps_score += (lcs_score * pst_score)
      else
        lcs_score = com_no1[com_cnt].length.to_f ** @len
        cps_score += lcs_score
      end
      com_cnt += 1
    end
    return cps_score
  end
    
  #
  #  set_tab
  #
  def set_tab(sent_no, com)
    com_cnt = 0
    while com_cnt < com.length
      pos_cnt = 0
      while pos_cnt < com[com_cnt].length
        pos = com[com_cnt][pos_cnt]
        @unit_sent[sent_no][pos] = "\t"
        if sent_no == 0
          i = 1
          while i < @list.length
            @list[i][pos+1] = "\t"
            i += 1
          end
        else
          i = 1
          while i < @list[0].length
            @list[pos+1][i] = "\t"
            i += 1
          end
        end
        pos_cnt += 1
      end
      com_cnt += 1
    end
  end
    
  #
  #  create_lcs_node_v2
  #
  def create_lcs_node_v2(no1, no2, com_edit, com_lcs, lcs, tree_edit, tree_lcs)
    ary = [com_lcs, com_edit].transpose
    
    ary.sort!
    
    com_cnt = 0
    while ary.length > com_cnt
      com_lcs[com_cnt] = ary[com_cnt][0]
      com_cnt += 1
    end
    
    com_cnt = 0
    while ary.length > com_cnt
      com_edit[com_cnt] = ary[com_cnt][1]
      com_cnt += 1
    end

    tmp_tree_edit = Array.new
    tmp_tree_lcs = Array.new
    com_cnt = 0
    while com_edit.length > com_cnt
      tmp_tree_edit[com_cnt] = com_edit[com_cnt]
      com_cnt += 1
    end
    com_cnt = 0
    while com_lcs.length > com_cnt
      tmp_tree_lcs[com_cnt] = com_lcs[com_cnt]
      com_cnt += 1
    end
    com_cnt = 0
    while com_edit.length > com_cnt
      tree_edit[com_cnt] = com_edit[com_cnt]
      com_cnt += 1
    end
    com_cnt = 0
    while com_lcs.length > com_cnt
      tree_lcs[com_cnt] = com_lcs[com_cnt]
      com_cnt += 1
    end
    
    tree_struct = Array.new
    
    loop do
      node_cnt = 0
      while tree_edit.length > node_cnt
        tree_struct[node_cnt] = []
        if lcs == tree_lcs[node_cnt]
          tree_struct[node_cnt] = [-1]
          node_cnt += 1
          next
        end
#=begin
        if @unit_sent[no1].length-tree_edit[node_cnt][1] <= lcs-tree_lcs[node_cnt] || @unit_sent[no2].length-tree_edit[node_cnt][0] <= lcs-tree_lcs[node_cnt]
          tree_struct[node_cnt] = []
          node_cnt += 1
          next
        end
#=end
        wk_node_cnt = node_cnt+1
        parent_cnt = 0
        while tree_edit.length > wk_node_cnt
          if tree_lcs[node_cnt]+1 >= tree_lcs[wk_node_cnt]
            if tree_lcs[node_cnt]+1 == tree_lcs[wk_node_cnt]
              if tree_edit[wk_node_cnt][0] > tree_edit[node_cnt][0] && tree_edit[wk_node_cnt][1] > tree_edit[node_cnt][1]
                tree_struct[node_cnt][parent_cnt] = wk_node_cnt
                parent_cnt += 1
              end
            end
            wk_node_cnt += 1
          else
            break
          end
        end
        node_cnt += 1
      end
      node_cnt = 0
      while tree_edit.length > node_cnt
        if tree_struct[node_cnt] == []
          del_par = node_cnt
          wk_node_cnt = node_cnt-1
          while wk_node_cnt >= 0
            if tree_lcs[node_cnt]-1 <= tree_lcs[wk_node_cnt]
              if tree_lcs[node_cnt]-1 == tree_lcs[wk_node_cnt]
                par_cnt = 0
                while tree_struct[wk_node_cnt].length > par_cnt
                  if tree_struct[wk_node_cnt][par_cnt] == del_par
                    tree_struct[wk_node_cnt].delete(del_par)
                    break
                  end
                  par_cnt += 1
                end
              end
            else
              break
            end
            wk_node_cnt -= 1
          end
        end
        node_cnt += 1
      end
#=begin
      node_cnt = tree_edit.length-1
      while node_cnt >= 0
        if tree_struct[node_cnt] == [-1]
          last_flg = 0
          wk_node_cnt = node_cnt-1
          while wk_node_cnt >= 0
            if tree_lcs[node_cnt]-1 <= tree_lcs[wk_node_cnt]
              if tree_lcs[node_cnt]-1 == tree_lcs[wk_node_cnt]
                par_cnt = 0
                while tree_struct[wk_node_cnt].length > par_cnt
                  if tree_struct[wk_node_cnt][par_cnt] == node_cnt
                    last_flg = 1
                    break
                  end
                  par_cnt += 1
                end
              end
            else
              if last_flg == 0
                tree_edit[node_cnt] = []
                tree_lcs[node_cnt] = []
              end
              break
            end
            if last_flg == 1
              break
            end
            wk_node_cnt -= 1
          end
        else
          break
        end
        node_cnt -= 1
      end
#=end

      node_cnt = 0
      while tree_edit.length > node_cnt
        if tree_struct[node_cnt] == []
          tree_edit[node_cnt] = []
          tree_lcs[node_cnt] = []
        end
        node_cnt += 1
      end
      tree_edit.delete([])
      tree_lcs.delete([])
      if tree_edit.length == tmp_tree_edit.length
        tree_edit.reverse!
        tree_lcs.reverse!
        break
      else
        tmp_tree_edit = []
        com_cnt = 0
        while tree_edit.length > com_cnt
          tmp_tree_edit[com_cnt] = tree_edit[com_cnt]
          com_cnt += 1
        end
        tmp_tree_lcs = []
        com_cnt = 0
        while tree_lcs.length > com_cnt
          tmp_tree_lcs[com_cnt] = tree_lcs[com_cnt]
          com_cnt += 1
        end
      end
    end
  end

  #
  #  continue_check
  #
  def continue_check(word, one_lcs_node, out_lcs_node, out_cnt, flg)
    wd_cnt = 0
    while one_lcs_node.length > wd_cnt
      word2 = one_lcs_node[wd_cnt]
      word_flg = 0
      if flg == 0
        if word[0]-1 == word2[0] && word[1]-1 == word2[1]
          word_flg = 1
        end
      end
      if flg == 1
        if word[0]+1 == word2[0] && word[1]+1 == word2[1]
          word_flg = 1
        end
      end
      if flg == 2
        if word[0] > word2[0] && word[1] > word2[1]
          word_flg = 1
        end
      end
      if flg == 3
        if word[0] < word2[0] && word[1] < word2[1]
          word_flg = 1
        end
      end
      if word_flg == 1
        out_lcs_node[out_cnt] = word
        out_cnt += 1
        break
      end
      wd_cnt += 1
    end
    return out_cnt
  end

  #
  #  edit_lcs_node
  #
  def edit_lcs_node(no1, no2, tree_edit, tree_lcs, out_edit, out_lcs)
    lcs_node = Array.new
    lcs_node = [[]]
    tree_cnt = 0
    lcs_no = @lcs
    lcs_cnt = 0
    while tree_lcs.length > tree_cnt
      if tree_lcs[tree_cnt] == lcs_no
        lcs_node[@lcs-lcs_no][lcs_cnt] = tree_edit[tree_cnt]
        lcs_cnt += 1
      else
        lcs_cnt = 0
        lcs_no -= 1
        lcs_node[@lcs-lcs_no] = []
        lcs_node[@lcs-lcs_no][lcs_cnt] = tree_edit[tree_cnt]
        lcs_cnt += 1
      end
      tree_cnt += 1
    end
    
    wk_lcs_node = lcs_node.dup
    lcs_cnt = 0
    while lcs_node.length > lcs_cnt
      if lcs_node[lcs_cnt].length == 1
        lcs_cnt += 1
        next
      end
      
      one_lcs_node = Array.new
      one_lcs_node = []
      one_cnt = 0
      wd_cnt = 0
      min = 1.0
      while lcs_node[lcs_cnt].length > wd_cnt
        word = lcs_node[lcs_cnt][wd_cnt]
        left = word[0]
        right = word[1]
        if lcs_cnt < @lcs-1
          one_cnt = continue_check(word, lcs_node[lcs_cnt+1], one_lcs_node, one_cnt, 0)
        end
        if lcs_cnt > 0
          one_cnt = continue_check(word, lcs_node[lcs_cnt-1], one_lcs_node, one_cnt, 1)
        end
        
        wd_pos_diff = ((left+1).to_f/@unit_sent[no2].length)-((right+1).to_f/@unit_sent[no1].length)
        wd_pos_diff = wd_pos_diff.abs
        if min > wd_pos_diff
          min = wd_pos_diff
          min_wd = wd_cnt
        end
        
        wd_cnt += 1
      end
      
      one_lcs_node[one_cnt] = lcs_node[lcs_cnt][min_wd]
      
      if one_lcs_node.length > 1
        one_lcs_node.uniq!
        lcs_node[lcs_cnt] = one_lcs_node
        one_lcs_node = []
        one_cnt = 0
      end
      lcs_cnt += 1
    end
    
    if lcs_node == wk_lcs_node
      tree_cnt = 0
      while tree_edit.length > tree_cnt
        out_edit[tree_cnt] = tree_edit[tree_cnt]
        out_lcs[tree_cnt] = tree_lcs[tree_cnt]
        tree_cnt += 1
      end
      return
    end
    
    loop do 
      wk_lcs_node = lcs_node.dup
      lcs_cnt = 0
      while lcs_node.length > lcs_cnt
        if lcs_node[lcs_cnt].length == 1
          lcs_cnt += 1
          next
        end
        one_lcs_node = []
        one_cnt = 0
        wd_cnt = 0
        while lcs_node[lcs_cnt].length > wd_cnt
          word = lcs_node[lcs_cnt][wd_cnt]
          left = word[0]
          right = word[1]
          if lcs_cnt < @lcs-1
            one_cnt = continue_check(word, lcs_node[lcs_cnt+1], one_lcs_node, one_cnt, 2)
          end
          if lcs_cnt == @lcs-1
            one_cnt = continue_check(word, lcs_node[lcs_cnt-1], one_lcs_node, one_cnt, 3)
          end
          wd_cnt += 1
        end
        if lcs_node[lcs_cnt] != one_lcs_node
          lcs_node[lcs_cnt] = one_lcs_node
          one_lcs_node = []
        end
        lcs_cnt += 1
      end
      if lcs_node == wk_lcs_node
        break
      end
    end

    lcs_cnt = 0
    tree_cnt = 0
    while lcs_node.length > lcs_cnt
      one_lcs_cnt = 0
      while lcs_node[lcs_cnt].length > one_lcs_cnt
        out_edit[tree_cnt] = lcs_node[lcs_cnt][one_lcs_cnt]
        out_lcs[tree_cnt] = @lcs-lcs_cnt
        tree_cnt += 1
        one_lcs_cnt += 1
      end
      lcs_cnt += 1
    end
  end
  
  #
  #  create_tree
  #
  def create_tree(tree_edit, tree_lcs, root_tree, tree_struct)
    edit_cnt = 0
    root_cnt = 0
    while tree_edit.length > edit_cnt
      if tree_lcs[edit_cnt] == @lcs
        root_tree[root_cnt] = edit_cnt
        root_cnt += 1
      end
      edit_cnt += 1
    end
    
    edit_cnt = 0
    while tree_edit.length > edit_cnt
      if tree_edit.length <= edit_cnt+1 || tree_lcs[edit_cnt] == 1
        tree_struct[edit_cnt] = []
        edit_cnt += 1
        next
      end
      tree_struct[edit_cnt] = []
      wk_edit_cnt = edit_cnt+1
      ary1 = tree_edit[edit_cnt]
      while tree_edit.length > wk_edit_cnt
        ary2 = tree_edit[wk_edit_cnt]
        if tree_lcs[edit_cnt] == tree_lcs[wk_edit_cnt]+1
          if ary1[0] > ary2[0] && ary1[1] > ary2[1]
            tree_struct[edit_cnt] << wk_edit_cnt
          end
        end
        wk_edit_cnt += 1
      end
      edit_cnt += 1
    end
  end

  #
  #  check_both_pos
  #
  def check_both_pos(com_no1, com_no2)
    pos_score = 0
    com_cnt = 0
    while com_cnt < com_no1.length
      pos_s_no1 = com_no1[com_cnt][0]
      pos_e_no1 = com_no1[com_cnt][com_no1[com_cnt].length-1]
      pos_s_no2 = com_no2[com_cnt][0]
      pos_e_no2 = com_no2[com_cnt][com_no2[com_cnt].length-1]
      pos_score += (pos_s_no1+pos_s_no2) + (pos_e_no1+pos_e_no2)
      com_cnt += 1
    end
    return pos_score.to_f / com_cnt
  end
  
  #
  #  check_ref_pos
  #
  def check_ref_pos(com_no)
    pos_score = 0
    com_cnt = 0
    while com_cnt <com_no.length
      pos_s = com_no[com_cnt][0]
      pos_e = com_no[com_cnt][com_no[com_cnt].length-1]
      pos_score += (pos_s+pos_e)
      com_cnt += 1
    end
    return pos_score.to_f / com_cnt
  end
  
  #
  #  lcs
  #
  def lcs(no1, no2, com_edit, com_lcs)
    str1 = @unit_sent[no1].dup
    str2 = @unit_sent[no2].dup
    
    com_cnt = 0
    i = 1
    while i < str2.length+1
      j = 1
      while j < str1.length+1
        if str1[j-1] == str2[i-1]
          @list[i][j] = @list[i-1][j-1]+1
          com_edit[com_cnt] = [i-1, j-1]
          com_lcs[com_cnt] = @list[i][j]
          com_cnt += 1
        else
          if @list[i-1][j] > @list[i][j-1]
            @list[i][j] = @list[i-1][j]
          else
            @list[i][j] = @list[i][j-1]
          end
        end
        j += 1
      end
      i += 1
    end
    return @list[str2.length][str1.length]
  end
  
  #
  #  rep_lcs
  #
  def rep_lcs(no1, no2, com_edit, com_lcs)
    str1 = @unit_sent[no1].dup
    str2 = @unit_sent[no2].dup
    
    com_cnt = 0
    i = 0
    while i < str2.length
      j = 0
      while j < str1.length
        if str1[j] == str2[i] && str1[j] != "\t" && str2[i] != "\t"
          com_edit[com_cnt] = [i, j]
          com_lcs[com_cnt] = @list[i+1][j+1]
          com_cnt += 1
        end
        j += 1
      end
      i += 1
    end
    
    str1.delete("\t")
    str2.delete("\t")
    
    rep_list = [[]]
    i = 0
    while i < str2.length+1
      rep_list[i] = [0]
      i += 1
    end
    j = 1
    while j < str1.length+1
      rep_list[0][j] = 0
      j += 1
    end

    com_cnt = 0
    i = 1
    while i < str2.length+1
      j = 1
      while j < str1.length+1
        if str1[j-1] == str2[i-1]
          rep_list[i][j] = rep_list[i-1][j-1]+1
          com_lcs[com_cnt] = rep_list[i][j]
          com_cnt += 1
        else
          if rep_list[i-1][j] > rep_list[i][j-1]
            rep_list[i][j] = rep_list[i-1][j]
          else
            rep_list[i][j] = rep_list[i][j-1]
          end
        end
        j += 1
      end
      i += 1
    end
    max = rep_list[str2.length][str1.length]
    
    i = 1
    rep_i = 1
    while i < @unit_sent[no2].length+1
      if @unit_sent[no2][i-1] == "\t"
        i += 1
        next
      end
      j = 1
      rep_j = 1
      while j < @unit_sent[no1].length+1
        if @list[i][j] == "\t"
          j += 1
          next
        end
        @list[i][j] = rep_list[rep_i][rep_j]
        j += 1
        rep_j += 1
      end
      i += 1
      rep_i += 1
    end
    return max
  end
  
  #
  #  get_lcs_route
  #
  def get_lcs_route(no1, no2)
    com_edit = Array.new
    com_lcs = Array.new
    tree_edit = Array.new
    tree_lcs = Array.new
    root_tree = Array.new
    
    tree_struct = Array.new
    max_com_no1 = Array.new
    max_com_no2 = Array.new
    wk_tree_edit = Array.new
    wk_tree_lcs = Array.new
    str1 = @unit_sent[no1].dup
    str1.delete("\t")
    if @unit_sent[no1].length == str1.length
      @lcs = lcs(no1, no2, com_edit, com_lcs)
    else
      @lcs = rep_lcs(no1, no2, com_edit, com_lcs)
    end
    if @lcs > 0
#print("LCS=", @lcs,"\n")
      create_lcs_node_v2(no1, no2, com_edit, com_lcs, @lcs, tree_edit, tree_lcs)
      edit_lcs_node(no1, no2, tree_edit, tree_lcs, wk_tree_edit, wk_tree_lcs)
      tree_edit = wk_tree_edit.dup
      tree_lcs = wk_tree_lcs.dup
      create_tree(tree_edit, tree_lcs, root_tree, tree_struct)
      max_cps_score = 0.0
      @max_route_num = 0
      @route_num = 0
      max_decide_com_part(max_com_no1, max_com_no2, no1, no2, root_tree, tree_struct, tree_edit, max_cps_score)
#  print("Number of LCS route=", @route_num, "\n")

=begin
print("Route No.", @max_route_num, "\n")
      disp_lcs_route(max_com_no1, no1)
      disp_lcs_route(max_com_no2, no2)
=end
      new_cps_score = cal_cps_score(max_com_no1, max_com_no2, no1, no2, 2)
      set_tab(no1, max_com_no1)
      set_tab(no2, max_com_no2)
      return new_cps_score
    else
      return 0
    end
  end
  
  #
  # compare_route
  #
  def compare_route(max_com_no1, max_com_no2, no1, no2, lcs_route, max_cps_score)
    com_no1 = Array.new
    com_no2 = Array.new
    node_cnt = 0
    com_cnt = 0
    ary1 = lcs_route[node_cnt]
    i_1 = ary1[0]
    j_1 = ary1[1]
    com_no1[com_cnt] = [j_1]
    com_no2[com_cnt] = [i_1]
    node_cnt += 1
    while lcs_route.length > node_cnt
      ary2 = lcs_route[node_cnt]
      i_2 = ary2[0]
      j_2 = ary2[1]
      if i_1 + 1 == i_2 && j_1 + 1 == j_2
        com_no1[com_cnt] << j_2
        com_no2[com_cnt] << i_2
      else
        com_cnt += 1
        com_no1[com_cnt] = [j_2]
        com_no2[com_cnt] = [i_2]
      end
      node_cnt += 1
      i_1 = i_2
      j_1 = j_2
    end

=begin
    print("No.", @route_num, "\n")
    disp_lcs_route(com_no1, no1)
    disp_lcs_route(com_no2, no2)
=end
    cps_score = cal_cps_score(com_no1, com_no2, no1, no2, 1)
    cps_score = cps_score ** (1.0/@len)
    if cps_score > max_cps_score
      max_cps_score = cps_score
      cpy_com(max_com_no1, max_com_no2, com_no1, com_no2)
      @max_route_num = @route_num
      return max_cps_score
    end
    if cps_score == max_cps_score
      pos_route = check_both_pos(max_com_no1, max_com_no2)
      pos_i = check_both_pos(com_no1, com_no2)
      if pos_i < pos_route
        cpy_com(max_com_no1, max_com_no2, com_no1, com_no2)
        @max_route_num = @route_num
        return max_cps_score
      end
      if pos_i == pos_route
        pos_route = check_ref_pos(max_com_no1)
        pos_i = check_ref_pos(com_no1)
        if pos_i < pos_route
          cpy_com(max_com_no1, max_com_no2, com_no1, com_no2)
          @max_route_num = @route_num
          return max_cps_score
        end
      end
    end
    return max_cps_score
  end
  
  #
  # cpy_com
  #
  def cpy_com(max_com_no1, max_com_no2, com_no1, com_no2)
    com_cnt = 0
    while max_com_no1.length > com_cnt
      max_com_no1.delete_at(com_cnt)
    end
    com_cnt = 0
    while max_com_no2.length > com_cnt
      max_com_no2.delete_at(com_cnt)
    end
    com_cnt = 0
    while com_no1.length > com_cnt
      max_com_no1[com_cnt] = com_no1[com_cnt]
      com_cnt += 1
    end
    com_cnt = 0
    while com_no2.length > com_cnt
      max_com_no2[com_cnt] = com_no2[com_cnt]
      com_cnt += 1
    end
  end
  
  #
  # max_decide_com_part
  #
  def max_decide_com_part(max_com_no1, max_com_no2, no1, no2, root_tree, tree_struct, tree_edit, max_cps_score)
    max_cps_score = 0.0
    branch_num = Array.new
    node_cnt = 0
    while tree_struct.length > node_cnt
      if tree_struct[node_cnt].length > 1
        branch_num[node_cnt] = 0
      else
        branch_num[node_cnt] = -1
      end
      node_cnt += 1
    end

    root_cnt = 0
    route_pos = Array.new
    while root_tree.length > root_cnt
      route_pos = []
      route_pos[0] = root_tree[root_cnt]
      max_cps_score = extract_best_route(max_com_no1, max_com_no2, no1, no2, route_pos, tree_struct, tree_edit, branch_num, max_cps_score)

=begin
      if @route_num > @limit_route_num
        return
      end
=end
      root_cnt += 1
    end
  end
  
  #
  # extract_best_route
  #
  def extract_best_route(max_com_no1, max_com_no2, no1, no2, route_pos, tree_struct, tree_edit, branch_num, max_cps_score)
    pos = route_pos[0]
    
    branch_cnt = 0
    loop do
      if tree_struct[pos].length != 0
        branch_cnt = branch_num[pos]
        route_pos << tree_struct[pos][branch_cnt]
        if branch_num[pos] != -1
          branch_num[pos] += 1
        end
        pos = tree_struct[pos][branch_cnt]
      else

        lcs_route = Array.new
        lcs_route = [[]]
        node_cnt = 0
        while route_pos.length > node_cnt
          lcs_route[node_cnt] = tree_edit[route_pos[node_cnt]]
          node_cnt += 1
        end

        @route_num += 1
        
        lcs_route.reverse!
        max_cps_score = compare_route(max_com_no1, max_com_no2, no1, no2, lcs_route, max_cps_score)
        
=begin
        if @route_num > @limit_route_num
          return max_cps_score
        end
=end
        loop do
          route_pos.delete_at(route_pos.length-1)
          if route_pos.length == 0
            return max_cps_score
          end
          pos = route_pos[route_pos.length-1]
          if tree_struct[pos].length > branch_num[pos] && branch_num[pos] != -1
            break
          end
          if branch_num[pos] != -1
            branch_num[pos] = 0
          end
          branch_cnt = 0
        end
      end
    end
  end

  #
  # extract_route
  #
  def extract_route(max_com_no1, max_com_no2, no1, no2, tree_struct, tree_lcs, node_pos, route_pos, tree_edit, max_cps_score)
    if tree_struct[node_pos].length == 0
    
      @route_num += 1

      lcs_route = Array.new
      lcs_route = [[]]
      node_cnt = 0
      while route_pos.length > node_cnt
        lcs_route[node_cnt] = tree_edit[route_pos[node_cnt]]
        node_cnt += 1
      end
      
      lcs_route.reverse!
      max_cps_score = compare_route(max_com_no1, max_com_no2, no1, no2, lcs_route, max_cps_score)
      
      route_pos.delete_at(route_pos.length-1)
      return max_cps_score
    end
    
    pos_cnt = 0
    while tree_struct[node_pos].length > pos_cnt
      route_pos << tree_struct[node_pos][pos_cnt]
      tree_pos = tree_struct[node_pos][pos_cnt]
      
      max_cps_score = extract_route(max_com_no1, max_com_no2, no1, no2, tree_struct, tree_lcs, tree_pos, route_pos, tree_edit, max_cps_score)
      
      pos_cnt += 1
    end
    
    route_pos.delete_at(route_pos.length-1)
    return max_cps_score

  end
  
  #
  #  disp_lcs_route
  #
  def disp_lcs_route(com_no, sent_no)
    com_cnt = 0
    pos = 0
    while com_cnt < com_no.length
      pos_cnt = 0
      while pos_cnt < com_no[com_cnt].length
        com_pos = com_no[com_cnt][pos_cnt]
        if pos_cnt == 0
          while pos < com_pos
            if @unit_sent[sent_no][pos] == "\t"
              print("\t", " ")
            else
              print(@unit_origin_sent[sent_no][pos], " ")
            end
            pos += 1
          end
          print("[")
        end
        if @unit_sent[sent_no][pos] == "\t"
          print("\t")
        else
          print(@unit_origin_sent[sent_no][pos])
        end
        pos += 1
        if pos_cnt == com_no[com_cnt].length-1
          print("] ")
        else
          print(" ")
        end
        pos_cnt += 1
      end
      com_cnt += 1
    end
    while pos < @unit_origin_sent[sent_no].length
      if @unit_sent[sent_no][pos] == "\t"
        print("\t", " ")
      else
        print(@unit_origin_sent[sent_no][pos], " ")
      end
      pos += 1
    end
    print("\n")
  end
  
end

# learn object
learn = Learn.new(ARGV[0], ARGV[1], ARGV[2].to_f, ARGV[3], ARGV[4].to_f)
learn.impact_process
